package at.ac.univie.cs.swa.component.architectureabstraction.validation

import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ArchitectureAbstractionDSLPackage
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentDef
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.Expression
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.Transformation
import at.ac.univie.cs.swa.component.architectureabstraction.utils.ArchitectureAbstractionUtils
import at.ac.univie.cs.swa.component.architectureabstraction.utils.DSLEvaluator
import com.google.inject.Inject
import java.util.ArrayList
import java.util.HashMap
import java.util.HashSet
import java.util.LinkedList
import java.util.List
import java.util.Map
import java.util.Map$Entry
import java.util.Set
import org.eclipse.emf.common.util.EList
import org.eclipse.emf.ecore.EObject
import org.eclipse.uml2.uml.Classifier
import org.eclipse.uml2.uml.NamedElement
import org.eclipse.xtext.resource.IEObjectDescription
import org.eclipse.xtext.scoping.IScope
import org.eclipse.xtext.scoping.IScopeProvider
import org.eclipse.xtext.validation.Check
import org.eclipse.xtext.validation.CheckType
import at.ac.univie.cs.swa.component.architectureabstraction.utils.AssociationGeneratorEvaluator
import org.eclipse.xtext.EcoreUtil2
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.VirtualConnectorAnnotation
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.VirtualConnector
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentConnector

public class ArchitectureAbstractionDSLValidator extends AbstractArchitectureAbstractionDSLValidator {

	extension ArchitectureAbstractionUtils = new ArchitectureAbstractionUtils

	public static val String NO_OBJECTS_REFERENCED = "_==NO_OBJECTS_REFERENCED==_";

	public static val String FILE_DOES_NOT_EXIST = "_==FILE_DOES_NOT_EXIST==_";

	public static val String NO_OBJECTS_REFERENCED_FOR_REGEXP = "_==NO_OBJECTS_REFERENCED_FOR_REGEXP==_";

	public static val String COMPONENT_DOES_NOT_EXIST = "_==COMPONENT_DOES_NOT_EXIST==_";

	//	private static val String COMPONENT_MAY_NOT_HAVE_NESTED_CLASSIFIERS = "_==COMPONENT_MAY_NOT_HAVE_NESTED_CLASSIFIERS==_";
	public static val String UNUSED_CLASSIFIERS = "_==CLASSIFIERS_UNUSED==_";

	private static val String DOUBLE_USED_CLASSIFIER = "_==DOUBLE_USED_CLASSIFIER==_";

	private AssociationGeneratorEvaluator evaluator;

	@Inject
	private IScopeProvider scopeProvider;

	@Check(CheckType::NORMAL)
	def checkDoubleUsedClasses(Transformation transformation) {
		var EList<ComponentDef> components = transformation.getComponents();

		var Map<NamedElement, Set<ComponentDef>> elementCount = new HashMap<NamedElement, Set<ComponentDef>>();
		var Map<NamedElement, Set<ComponentDef>> multipleUsedElements = new HashMap<NamedElement, Set<ComponentDef>>();

		for (ComponentDef compDef : components) {
			var Set<Classifier> usedElements = getEvaluator().evaluateComponent(compDef);

			for (NamedElement element : usedElements) {
				if (elementCount.containsKey(element)) {
					var Set<ComponentDef> comps = elementCount.get(element);
					comps.add(compDef);

					elementCount.put(element, comps);

					multipleUsedElements.put(element, comps);

				} else {
					var Set<ComponentDef> comps = new HashSet<ComponentDef>();
					comps.add(compDef);
					elementCount.put(element, comps);
				}

			}

		}
		var Set<Entry<NamedElement, Set<ComponentDef>>> multipleUsedElmentEntrySet = multipleUsedElements.entrySet();
		for (Entry<NamedElement, Set<ComponentDef>> entry : multipleUsedElmentEntrySet) {
			var Set<ComponentDef> multipleUsedComponentList = entry.getValue();

			if (multipleUsedComponentList.size() > 1) {
				var key = entry.getKey()
				if (key.eIsProxy) {
					EcoreUtil2::resolveAll(key)
				}
				var String warningMessage = "Multiple used classifier: " + key + " used in: ";
				var boolean start = true;
				for (ComponentDef compdef : multipleUsedComponentList) {
					if (!start) {
						warningMessage = warningMessage + ", ";
					} else {
						start = false;
					}
					warningMessage = warningMessage + compdef.getName();

				}

				for (ComponentDef compdef : multipleUsedComponentList) {
					error(warningMessage, compdef,
						ArchitectureAbstractionDSLPackage::eINSTANCE.getComponentDef_Expr(),
						ArchitectureAbstractionDSLValidator::DOUBLE_USED_CLASSIFIER, warningMessage);

				}
			}

		}

	}

	def private AssociationGeneratorEvaluator getEvaluator() {
		if (evaluator == null) {
			evaluator = new AssociationGeneratorEvaluator();
			evaluator.setScopeProvider(scopeProvider);
		}

		// if(evaluator.getScopeProvider() == null) {
		// evaluator.setScopeProvider(scopeProvider);
		// }
		return evaluator;
	}

	@Check(CheckType::EXPENSIVE)
	def public void checkUnusedExistingClasses(Transformation transformation) {
		var IScope scope = scopeProvider.getScope(transformation,
			ArchitectureAbstractionDSLPackage::eINSTANCE.getIsClass_RelatedTo());
		var EList<ComponentDef> componentList = transformation.getComponents();
		var Iterable<IEObjectDescription> allDescriptions = scope.getAllElements();
		var List<Classifier> unUsed = new LinkedList<Classifier>();
		for (IEObjectDescription desc : allDescriptions) {

			var EObject eObjectOrProxy = desc.getEObjectOrProxy();
			if (eObjectOrProxy instanceof Classifier) {
				var Classifier classifier = eObjectOrProxy as Classifier;
				unUsed.add(classifier);
			}

		}
		for (ComponentDef compDef : componentList) {
			unUsed.removeAll(getEvaluator().evaluateComponent(compDef));
		}

		var List<String> issueData = new ArrayList<String>();
		if (!unUsed.isEmpty()) {
			var String string = "The following classifiers are not used in any component description:";

			for (Classifier classifier : unUsed) {
				var String convertedName = convertName(classifier);
				if (convertedName != null) {
					string = string + "\n" + convertedName;
					issueData.add(convertedName);
				}
			}
			var String[] array;
			array = issueData.toArray(array);

			if (string.length() > 1024) {
				string = string.substring(0, 1024);
			}
			if (array.size > 100) {
				var String[] shortarray = new ArrayList<String>(100);
				var int count = 0;
				while (count < 100) {
					shortarray.set(count, array.get(count));
					count = count + 1;
				}

				//				for (i = 0; i < shortarray.size; i++) {
				//					shortarray[i] = array[i];
				//				}
				array = shortarray;
			}
			warning(string, transformation,
				ArchitectureAbstractionDSLPackage::eINSTANCE.getTransformation_Components(), UNUSED_CLASSIFIERS, array);

		}

	}

	def String convertName(Classifier classifier) {
		var String qualifiedName = classifier.getQualifiedName();
		var String prefix = "root::";
		var String classifierName = classifier.getName();
		if (qualifiedName == null && classifierName != null) {
			return classifierName;
		}
		if (qualifiedName != null && qualifiedName.startsWith(prefix)) {
			qualifiedName = qualifiedName.substring(prefix.length());
		}
		return qualifiedName;
	}

	@Check(CheckType::NORMAL)
	def doCheckTest(Expression expr) {
		var Set<Classifier> matchedObjects = getEvaluator().checkExpression(expr);
		if (matchedObjects.isEmpty()) {
			warning("No classes mapped by this expression: " + expr.eClass().getName(), expr, null,
				NO_OBJECTS_REFERENCED);
		}
	}

	@Check(CheckType::EXPENSIVE)
	def doCheckRelationHasSourceRelation(ComponentDef source) {
		val sourceRelatedComponentNames = calculatedDirectRelatedComponents(source, evaluator).map[e|e.name].toSet
		for (ca : source.connectors) {
			doCheckConnectorHasSourceRelation(ca, sourceRelatedComponentNames, source)

		} 
	}

	def doCheckConnectorHasSourceRelation(ComponentConnector ca, Set<String> sourceRelatedComponentNames, ComponentDef source) {
		if (!(ca instanceof VirtualConnectorAnnotation)) {

			val implementingExpression = ca.implementingExpression

			// only check for source relations if no explicit implementation of the connector is given
			if (!(implementingExpression?.size > 0 || ca.implementingRelations?.size > 0)) {

				for (target : ca.targetComponents) {
					if (!sourceRelatedComponentNames.contains(target.name)) {

						//					if (!source.sourceRelationExists(target)) {
						warning(
							'''In the source no relation from «source.name» to target «target.name» exists!''',
							ArchitectureAbstractionDSLPackage::eINSTANCE.componentDef_Connectors,
							source.connectors.indexOf(ca)
						);

					}
				}

			}
		}
	}

	//	@Check
	//	def doCheckIndirection(IndirectionAnnotation indirectionAnnotation) {
	//		val ComponentDef proxy = indirectionAnnotation.eContainer as ComponentDef
	//		val ComponentDef target = indirectionAnnotation.target
	//		
	//		/* not sure if 2 relations are necessary - 1 might be sufficient
	//		 * especially if  
	//		 */
	//		if(!sourceRelationExists(proxy,target)) {
	//			warning('''No relation from proxy component to target component'''
	//				, indirectionAnnotation,
	//				ArchitectureAbstractionDSLPackage::eINSTANCE.indirectionAnnotation_Target,0
	//			);
	//		}
	//		
	//		if(!sourceRelationExists(target,proxy)) {
	//			warning('''No relation from target component to proxy component'''
	//				, indirectionAnnotation,
	//				ArchitectureAbstractionDSLPackage::eINSTANCE.indirectionAnnotation_Target,0
	//			);
	//		}
	//	}
	def boolean relationExists(ComponentDef source, ComponentDef target) {
		if (directrelationExists(source, target)) {
			return true;
		} else {
			return relationExists(source, target, new HashSet<ComponentDef>())
		}
	}

	def boolean relationExists(ComponentDef source, ComponentDef target, Set<ComponentDef> visitedComponents) {
		visitedComponents.add(source)
		if (source.connectors.empty) {
			return false
		}
		return source.connectors.exists(
			e|e.targetComponents.filterNull.exists(f|f.relationExists(target, visitedComponents)))
	}

	def directrelationExists(ComponentDef source, ComponentDef target) {
		return source.connectors.contains(target)
	}

	//		def sourceRelationExists(ComponentDef source, ComponentDef target) {
	//			sourceRelationExists(source, target, false)
	//		}
	//
	//		def boolean sourceRelationExists(ComponentDef source, ComponentDef target, boolean print) {
	//			if (print) {
	//				System::out.println("sourceRelationExists - source: " + source.name + " target: " + target.name)
	//			}
	//			try {
	//				val proxyClassifiers = evaluator.evaluateComponent(source)
	//				if (print) {
	//					System::out.println("sourceClassifiers: " + proxyClassifiers)
	//				}
	//
	//				val targetClassifiers = evaluator.evaluateComponent(target)
	//				if (print) {
	//					System::out.println("targetClassifiers: " + targetClassifiers)
	//				}
	//				val targetClassifierNames = targetClassifiers.map(f|f.name).filterNull.toSet.sort
	//				if (print) {
	//					System::out.println("targetClassifierNames: " + targetClassifierNames)
	//				}
	//				val Set<Classifier> relTargetClassifiers = proxyClassifiers.map(c|c.relatedTargetClassifiers(false)).
	//					flatten.toSet
	//				val List<String> relTargetClassifierNames = relTargetClassifiers.map(c|c.name).sort.toList
	//				if (print) {
	//					System::out.println("relTargetClassifierNames: " + relTargetClassifierNames)
	//				}
	//
	//				//proxyClassifiers.exists(c | c.relatedTargetClassifiers(print).exists(e| targetClassifierNames.contains(e.name)))
	//				return relTargetClassifierNames.exists(e|targetClassifierNames.contains(e))
	//			} catch (Exception e) {
	//				e.printStackTrace
	//			}
	//			return false
	//
	//		}
	def Transformation getTransformation(EObject eObj) {
		return eObj.eResource.contents.first as Transformation
	}

	def first(List<EObject> list) {
		if (list.size > 0)
			return list.get(0)
		return null
	}

}
